#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <malloc.h>
#include "dynamic_libs/os_functions.h"
#include "dynamic_libs/fs_functions.h"
#include "dynamic_libs/gx2_functions.h"
#include "dynamic_libs/sys_functions.h"
#include "dynamic_libs/vpad_functions.h"
#include "dynamic_libs/padscore_functions.h"
#include "dynamic_libs/socket_functions.h"
#include "dynamic_libs/ax_functions.h"
#include "fs/fs_utils.h"
#include "fs/sd_fat_devoptab.h"
#include "system/memory.h"
#include "utils/logger.h"
#include "utils/utils.h"
#include "common/common.h"
#include "glmWrapper.h"
#include "shader.h"
#include "model.h"

#define ATTRIBUTE_COUNT                 2

static GX2VertexShader *vertexShader = NULL;
static GX2PixelShader *pixelShader = NULL;
static GX2FetchShader *fetchShader = NULL;
static GX2AttribStream *attributes = NULL;

static f32 * cubePosVtx = NULL;
static f32 * cubeColorBuf = NULL;

static Mtx44 projectionMtx;
static Mtx44 viewMtx;

static f32 degreeX = 0.0f;
static f32 degreeY = 0.0f;
static f32 degreeZ = 0.0f;
static bool manualControl = false;

static void setupScene(void)
{
    //! all GX2 variables have to reside inside valid memory range and can not be inside our own setup range (0x00800000 - 0x01000000)
    //! therefore we move all static data from the data section related to GX2 to heap memory

    //!----------------------------------------------------------------------------------------------------------
    //! setup vertex shader
    //!----------------------------------------------------------------------------------------------------------
    vertexShader = (GX2VertexShader*) memalign(0x40, sizeof(GX2VertexShader));
    memset(vertexShader, 0, sizeof(GX2VertexShader));
    vertexShader->shader_mode = GX2_SHADER_MODE_UNIFORM_REGISTER;
    vertexShader->shader_size = sizeof(ucVertexProgram);
    vertexShader->shader_data = memalign(GX2_SHADER_ALIGNMENT, vertexShader->shader_size);
    memcpy(vertexShader->shader_data, ucVertexProgram, vertexShader->shader_size);
    GX2Invalidate(GX2_INVALIDATE_CPU_SHADER, vertexShader->shader_data, vertexShader->shader_size);
    memcpy(vertexShader->regs, cuVertexProgramRegs, sizeof(cuVertexProgramRegs));

    vertexShader->uniform_vars_count = 1;
    vertexShader->uniform_var = (GX2UniformVar*) malloc(vertexShader->uniform_vars_count * sizeof(GX2UniformVar));
    vertexShader->uniform_var[0].name = "MVP";
    vertexShader->uniform_var[0].var_type = GX2_VAR_TYPE_MAT4;
    vertexShader->uniform_var[0].array_count = 1;
    vertexShader->uniform_var[0].offset = 0;
    vertexShader->uniform_var[0].block_index = -1;

    vertexShader->attribute_vars_count = 2;
    vertexShader->attribute_var = (GX2AttribVar*) malloc(vertexShader->attribute_vars_count * sizeof(GX2AttribVar));
    vertexShader->attribute_var[0].name = "vertexColor";
    vertexShader->attribute_var[0].var_type = GX2_VAR_TYPE_VEC3;
    vertexShader->attribute_var[0].array_count = 0;
    vertexShader->attribute_var[0].location = 1;
    vertexShader->attribute_var[1].name = "vertexPosition_modelspace";
    vertexShader->attribute_var[1].var_type = GX2_VAR_TYPE_VEC3;
    vertexShader->attribute_var[1].array_count = 0;
    vertexShader->attribute_var[1].location = 0;

    //!----------------------------------------------------------------------------------------------------------
    //! setup pixel shader
    //!----------------------------------------------------------------------------------------------------------
    pixelShader = (GX2PixelShader*) memalign(0x40, sizeof(GX2PixelShader));
    memset(pixelShader, 0, sizeof(GX2PixelShader));
    pixelShader->shader_mode = GX2_SHADER_MODE_UNIFORM_REGISTER;
    pixelShader->shader_size = sizeof(ucPixelProgram);
    pixelShader->shader_data = memalign(GX2_SHADER_ALIGNMENT, pixelShader->shader_size);
    memcpy(pixelShader->shader_data, ucPixelProgram, pixelShader->shader_size);
    GX2Invalidate(GX2_INVALIDATE_CPU_SHADER, pixelShader->shader_data, pixelShader->shader_size);
    memcpy(pixelShader->regs, ucPixelProgrammRegs, sizeof(ucPixelProgrammRegs));

    //!----------------------------------------------------------------------------------------------------------
    //! setup attributes
    //!----------------------------------------------------------------------------------------------------------
    attributes = (GX2AttribStream*) malloc(sizeof(GX2AttribStream) * ATTRIBUTE_COUNT);
    GX2InitAttribStream(&attributes[0], vertexShader->attribute_var[1].location, 0, 0, GX2_ATTRIB_FORMAT_32_32_32_FLOAT);
    GX2InitAttribStream(&attributes[1], vertexShader->attribute_var[0].location, 1, 0, GX2_ATTRIB_FORMAT_32_32_32_FLOAT);

    //!----------------------------------------------------------------------------------------------------------
    //! setup fetch shader for the attributes
    //!----------------------------------------------------------------------------------------------------------
    u32 shaderSize = GX2CalcFetchShaderSizeEx(ATTRIBUTE_COUNT, GX2_FETCH_SHADER_TESSELATION_NONE, GX2_TESSELLATION_MODE_DISCRETE);
    void *fetchShaderProgramm = memalign(GX2_SHADER_ALIGNMENT, shaderSize);
    fetchShader = (GX2FetchShader *) malloc(sizeof(GX2FetchShader));
    GX2InitFetchShaderEx(fetchShader, fetchShaderProgramm, ATTRIBUTE_COUNT, attributes, GX2_FETCH_SHADER_TESSELATION_NONE, GX2_TESSELLATION_MODE_DISCRETE);
    GX2Invalidate(GX2_INVALIDATE_CPU_SHADER, fetchShaderProgramm, shaderSize);

    //!----------------------------------------------------------------------------------------------------------
    //! move cube vertex and color buffer to heap
    //!----------------------------------------------------------------------------------------------------------
    cubePosVtx = (f32 *)memalign(0x40, sizeof(g_vertex_buffer_data));
    memcpy(cubePosVtx, g_vertex_buffer_data, sizeof(g_vertex_buffer_data));
    GX2Invalidate(GX2_INVALIDATE_CPU_ATTRIB_BUFFER, cubePosVtx, sizeof(g_vertex_buffer_data));

    cubeColorBuf = (f32 *)memalign(0x40, sizeof(g_color_buffer_data));
    memcpy(cubeColorBuf, g_color_buffer_data, sizeof(g_color_buffer_data));
    GX2Invalidate(GX2_INVALIDATE_CPU_ATTRIB_BUFFER, cubeColorBuf, sizeof(g_color_buffer_data));
}

static void prepareRendering(GX2ColorBuffer * currColorBuffer, GX2DepthBuffer * currDepthBuffer, GX2ContextState * currContextState)
{
    GX2ClearColor(currColorBuffer, 0.0f, 0.0f, 0.4f, 1.0f);
    GX2ClearDepthStencilEx(currDepthBuffer, currDepthBuffer->clear_depth, currDepthBuffer->clear_stencil, GX2_CLEAR_BOTH);

    GX2SetContextState(currContextState);
    GX2SetViewport(0.0f, 0.0f, currColorBuffer->surface.width, currColorBuffer->surface.height, 0.0f, 1.0f);
    GX2SetScissor(0, 0, currColorBuffer->surface.width, currColorBuffer->surface.height);

    GX2SetDepthOnlyControl(GX2_ENABLE, GX2_ENABLE, GX2_COMPARE_LESS);
    GX2SetColorControl(GX2_LOGIC_OP_COPY, 1, GX2_DISABLE, GX2_ENABLE);
    GX2SetBlendControl(GX2_RENDER_TARGET_0, GX2_BLEND_SRC_ALPHA, GX2_BLEND_ONE_MINUS_SRC_ALPHA, GX2_BLEND_COMBINE_ADD, GX2_ENABLE, GX2_BLEND_SRC_ALPHA, GX2_BLEND_ONE_MINUS_SRC_ALPHA, GX2_BLEND_COMBINE_ADD);
    GX2SetCullOnlyControl(GX2_FRONT_FACE_CCW, GX2_DISABLE, GX2_DISABLE);
}

static void renderScene(void)
{
    GX2SetFetchShader(fetchShader);
    GX2SetVertexShader(vertexShader);
    GX2SetPixelShader(pixelShader);

    GX2SetAttribBuffer(0, sizeof(g_vertex_buffer_data), sizeof(f32) * 3, cubePosVtx);
    GX2SetAttribBuffer(1, sizeof(g_color_buffer_data), sizeof(f32) * 3, cubeColorBuf);

    Mtx44 modelMtx;
	glmIdentity(modelMtx);
	glmRotate(modelMtx, DegToRad(degreeX), 1.0f, 0.0f, 0.0f);
    glmRotate(modelMtx, DegToRad(degreeY), 0.0f, 1.0f, 0.0f);
	glmRotate(modelMtx, DegToRad(degreeZ), 0.0f, 0.0f, 1.0f);

    if(!manualControl)
    {
        degreeX += 0.5f;
        degreeY += 0.8f;
        degreeZ += 0.4f;
    }

    Mtx44 mtxMVP;
	glmMultiply(mtxMVP, projectionMtx, viewMtx);
	glmMultiply(mtxMVP, mtxMVP, modelMtx);

    GX2SetVertexUniformReg(vertexShader->uniform_var[0].offset, 16, &mtxMVP[0][0]);

    GX2DrawEx(GX2_PRIMITIVE_TRIANGLES, vtxCount, 0, 1);
}

/* Entry point */
int Menu_Main(void)
{
    //!*******************************************************************
    //!                   Initialize function pointers                   *
    //!*******************************************************************
    //! do OS (for acquire) and sockets first so we got logging
    InitOSFunctionPointers();
    InitSocketFunctionPointers();
    InitVPadFunctionPointers();
    InitGX2FunctionPointers();

    //!*******************************************************************
    //!                    Initialize heap memory                        *
    //!*******************************************************************
    memoryInitialize();

    //!*******************************************************************
    //!                        Initialize FS                             *
    //!*******************************************************************
    //! allocate MEM2 command buffer memory
    unsigned char *gx2CommandBuffer = MEM2_alloc(GX2_COMMAND_BUFFER_SIZE, 0x40);

    //! initialize GX2 command buffer
    u32 gx2_init_attributes[9];
    gx2_init_attributes[0] = GX2_INIT_ATTRIB_CB_BASE;
    gx2_init_attributes[1] = (u32)gx2CommandBuffer;
    gx2_init_attributes[2] = GX2_INIT_ATTRIB_CB_SIZE;
    gx2_init_attributes[3] = GX2_COMMAND_BUFFER_SIZE;
    gx2_init_attributes[4] = GX2_INIT_ATTRIB_ARGC;
    gx2_init_attributes[5] = 0;
    gx2_init_attributes[6] = GX2_INIT_ATTRIB_ARGV;
    gx2_init_attributes[7] = 0;
    gx2_init_attributes[8] = GX2_INIT_ATTRIB_NULL;
    GX2Init(gx2_init_attributes);

    u32 scanBufferSize = 0;
    s32 scaleNeeded = 0;

    s32 tvScanMode = GX2GetSystemTVScanMode();
    s32 drcScanMode = GX2GetSystemDRCScanMode();

    s32 tvRenderMode;
    u32 tvWidth = 0;
    u32 tvHeight = 0;

    switch(tvScanMode)
    {
    case GX2_TV_SCAN_MODE_480I:
    case GX2_TV_SCAN_MODE_480P:
        tvWidth = 854;
        tvHeight = 480;
        tvRenderMode = GX2_TV_RENDER_480_WIDE;
        break;
    case GX2_TV_SCAN_MODE_1080I:
    case GX2_TV_SCAN_MODE_1080P:
        tvWidth = 1920;
        tvHeight = 1080;
        tvRenderMode = GX2_TV_RENDER_1080;
        break;
    case GX2_TV_SCAN_MODE_720P:
    default:
        tvWidth = 1280;
        tvHeight = 720;
        tvRenderMode = GX2_TV_RENDER_720;
        break;
    }

    s32 tvAAMode = GX2_AA_MODE_1X;
    s32 drcAAMode = GX2_AA_MODE_4X;

    //! calculate the size needed for the TV scan buffer and allocate the buffer from bucket memory
    GX2CalcTVSize(tvRenderMode, GX2_SURFACE_FORMAT_TCS_R8_G8_B8_A8_UNORM, GX2_BUFFERING_DOUBLE, &scanBufferSize, &scaleNeeded);
    unsigned char *tvScanBuffer = MEMBucket_alloc(scanBufferSize, GX2_SCAN_BUFFER_ALIGNMENT);
    GX2Invalidate(GX2_INVALIDATE_CPU, tvScanBuffer, scanBufferSize);
    GX2SetTVBuffer(tvScanBuffer, scanBufferSize, tvRenderMode, GX2_SURFACE_FORMAT_TCS_R8_G8_B8_A8_UNORM, GX2_BUFFERING_DOUBLE);

    //! calculate the size needed for the DRC scan buffer and allocate the buffer from bucket memory
    GX2CalcDRCSize(drcScanMode, GX2_SURFACE_FORMAT_TCS_R8_G8_B8_A8_UNORM, GX2_BUFFERING_DOUBLE, &scanBufferSize, &scaleNeeded);
    unsigned char *drcScanBuffer = MEMBucket_alloc(scanBufferSize, GX2_SCAN_BUFFER_ALIGNMENT);
    GX2Invalidate(GX2_INVALIDATE_CPU, drcScanBuffer, scanBufferSize);
    GX2SetDRCBuffer(drcScanBuffer, scanBufferSize, drcScanMode, GX2_SURFACE_FORMAT_TCS_R8_G8_B8_A8_UNORM, GX2_BUFFERING_DOUBLE);

    //! Setup color buffer for TV rendering
    GX2ColorBuffer tvColorBuffer;
    GX2InitColorBuffer(&tvColorBuffer, GX2_SURFACE_DIM_2D, tvWidth, tvHeight, 1, GX2_SURFACE_FORMAT_TCS_R8_G8_B8_A8_UNORM, tvAAMode);
    tvColorBuffer.surface.image_data = MEM1_alloc(tvColorBuffer.surface.image_size, tvColorBuffer.surface.align);
    GX2Invalidate(GX2_INVALIDATE_CPU, tvColorBuffer.surface.image_data, tvColorBuffer.surface.image_size);

    //! Setup TV depth buffer (can be the same for both if rendered one after another)
    u32 size, align;
    GX2DepthBuffer tvDepthBuffer;
    GX2InitDepthBuffer(&tvDepthBuffer, GX2_SURFACE_DIM_2D, tvColorBuffer.surface.width, tvColorBuffer.surface.height, 1, GX2_SURFACE_FORMAT_TCD_R32_FLOAT, tvAAMode);
    tvDepthBuffer.surface.image_data = MEM1_alloc(tvDepthBuffer.surface.image_size, tvDepthBuffer.surface.align);
    GX2Invalidate(GX2_INVALIDATE_CPU, tvDepthBuffer.surface.image_data, tvDepthBuffer.surface.image_size);

    //! Setup TV HiZ buffer
    GX2CalcDepthBufferHiZInfo(&tvDepthBuffer, &size, &align);
    tvDepthBuffer.hiZ_data = MEM1_alloc(size, align);
    GX2Invalidate(GX2_INVALIDATE_CPU, tvDepthBuffer.hiZ_data, size);
    GX2InitDepthBufferHiZEnable(&tvDepthBuffer, GX2_ENABLE);

    //! Setup color buffer for DRC rendering
    GX2ColorBuffer drcColorBuffer;
    GX2InitColorBuffer(&drcColorBuffer, GX2_SURFACE_DIM_2D, 854, 480, 1, GX2_SURFACE_FORMAT_TCS_R8_G8_B8_A8_UNORM, drcAAMode);
    drcColorBuffer.surface.image_data = MEM1_alloc(drcColorBuffer.surface.image_size, drcColorBuffer.surface.align);
    GX2Invalidate(GX2_INVALIDATE_CPU, drcColorBuffer.surface.image_data, drcColorBuffer.surface.image_size);

    //! Setup DRC depth buffer (can be the same for both if rendered one after another)
    GX2DepthBuffer drcDepthBuffer;
    GX2InitDepthBuffer(&drcDepthBuffer, GX2_SURFACE_DIM_2D, drcColorBuffer.surface.width, drcColorBuffer.surface.height, 1, GX2_SURFACE_FORMAT_TCD_R32_FLOAT, drcAAMode);
    drcDepthBuffer.surface.image_data = MEM1_alloc(drcDepthBuffer.surface.image_size, drcDepthBuffer.surface.align);
    GX2Invalidate(GX2_INVALIDATE_CPU, drcDepthBuffer.surface.image_data, drcDepthBuffer.surface.image_size);

    //! Setup DRC HiZ buffer
    GX2CalcDepthBufferHiZInfo(&drcDepthBuffer, &size, &align);
    drcDepthBuffer.hiZ_data = MEM1_alloc(size, align);
    GX2Invalidate(GX2_INVALIDATE_CPU, drcDepthBuffer.hiZ_data, size);
    GX2InitDepthBufferHiZEnable(&drcDepthBuffer, GX2_ENABLE);


    //! allocate auxilary buffer last as there might not be enough MEM1 left for other stuff after that
    if (tvColorBuffer.surface.aa)
    {
        u32 auxSize, auxAlign;
        GX2CalcColorBufferAuxInfo(&tvColorBuffer, &auxSize, &auxAlign);
        tvColorBuffer.aux_data = MEM1_alloc(auxSize, auxAlign);
        if(!tvColorBuffer.aux_data)
            tvColorBuffer.aux_data = MEM2_alloc(auxSize, auxAlign);

        tvColorBuffer.aux_size = auxSize;
        memset(tvColorBuffer.aux_data, GX2_AUX_BUFFER_CLEAR_VALUE, auxSize);
        GX2Invalidate(GX2_INVALIDATE_CPU, tvColorBuffer.aux_data, auxSize);
    }

    if (drcColorBuffer.surface.aa)
    {
        u32 auxSize, auxAlign;
        GX2CalcColorBufferAuxInfo(&drcColorBuffer, &auxSize, &auxAlign);
        drcColorBuffer.aux_data = MEM1_alloc(auxSize, auxAlign);
        if(!drcColorBuffer.aux_data)
            drcColorBuffer.aux_data = MEM2_alloc(auxSize, auxAlign);
        drcColorBuffer.aux_size = auxSize;
        memset(drcColorBuffer.aux_data, GX2_AUX_BUFFER_CLEAR_VALUE, auxSize);
        GX2Invalidate(GX2_INVALIDATE_CPU, drcColorBuffer.aux_data, auxSize );
    }

    //! allocate memory and setup context state TV
    GX2ContextState *tvContextState = (GX2ContextState*)MEM2_alloc(sizeof(GX2ContextState), GX2_CONTEXT_STATE_ALIGNMENT);
    GX2SetupContextStateEx(tvContextState, GX2_TRUE);

    //! allocate memory and setup context state DRC
    GX2ContextState *drcContextState = (GX2ContextState*)MEM2_alloc(sizeof(GX2ContextState), GX2_CONTEXT_STATE_ALIGNMENT);
    GX2SetupContextStateEx(drcContextState, GX2_TRUE);

    //! set initial context state and render buffers
    GX2SetContextState(tvContextState);
    GX2SetColorBuffer(&tvColorBuffer, GX2_RENDER_TARGET_0);
    GX2SetDepthBuffer(&tvDepthBuffer);

    GX2SetContextState(drcContextState);
    GX2SetColorBuffer(&drcColorBuffer, GX2_RENDER_TARGET_0);
    GX2SetDepthBuffer(&drcDepthBuffer);

	glmPerspective(projectionMtx, 45.0f, 16.0f / 9.0f, 0.1f, 100.0f);
	glmLookAt(viewMtx, 4.0f, 3.0f, -3.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);

    setupScene();

    int vpadError = -1;
    VPADData vpad;

    int firstFrameDone = 0;

    while(1)
    {
        VPADRead(0, &vpad, 1, &vpadError);

        if(vpadError == 0 && ((vpad.btns_d | vpad.btns_h) & VPAD_BUTTON_HOME))
            break;

        if(vpad.btns_h & VPAD_BUTTON_LEFT) {
            degreeX -= 1.0f;
            manualControl = true;
        }
        if(vpad.btns_h & VPAD_BUTTON_RIGHT) {
            degreeX += 1.0f;
            manualControl = true;
        }
        if(vpad.btns_h & VPAD_BUTTON_UP) {
            degreeY -= 1.0f;
            manualControl = true;
        }
        if(vpad.btns_h & VPAD_BUTTON_DOWN) {
            degreeY += 1.0f;
            manualControl = true;
        }
        if(vpad.btns_h & VPAD_BUTTON_L) {
            degreeZ -= 1.0f;
            manualControl = true;
        }
        if(vpad.btns_h & VPAD_BUTTON_R) {
            degreeZ += 1.0f;
            manualControl = true;
        }
        if(vpad.btns_h & VPAD_BUTTON_A) {
            manualControl = false;
        }


        //!*******************************************************************
        //!                          DRC Rending                             *
        //!*******************************************************************
        prepareRendering(&drcColorBuffer, &drcDepthBuffer, drcContextState);
        renderScene();
        GX2CopyColorBufferToScanBuffer(&drcColorBuffer, GX2_SCAN_TARGET_DRC_FIRST);

        //!*******************************************************************
        //!                          TV Rending                              *
        //!*******************************************************************
        prepareRendering(&tvColorBuffer, &tvDepthBuffer, tvContextState);
        renderScene();
        GX2CopyColorBufferToScanBuffer(&tvColorBuffer, GX2_SCAN_TARGET_TV);
        GX2SwapScanBuffers();
        GX2Flush();

        if(!firstFrameDone)
        {
            firstFrameDone = 1;
            GX2SetTVEnable(GX2_TRUE);
            GX2SetDRCEnable(GX2_TRUE);
        }

        GX2WaitForVsync();
    }

    GX2SetTVEnable(GX2_FALSE);
    GX2SetDRCEnable(GX2_FALSE);
    GX2WaitForVsync();
    GX2DrawDone();
    GX2Shutdown();

    //!*******************************************************************
    //!                        Let's clean up                            *
    //!*******************************************************************
    MEM2_free(gx2CommandBuffer);
    MEMBucket_free(tvScanBuffer);
    MEMBucket_free(drcScanBuffer);
    MEM1_free(tvColorBuffer.surface.image_data);
    MEM1_free(tvDepthBuffer.surface.image_data);
    MEM1_free(tvDepthBuffer.hiZ_data);
    MEM1_free(drcColorBuffer.surface.image_data);
    MEM1_free(drcDepthBuffer.surface.image_data);
    MEM1_free(drcDepthBuffer.hiZ_data);
    MEM2_free(tvContextState);
    MEM2_free(drcContextState);

    free(vertexShader->uniform_var);
    free(vertexShader->attribute_var);
    free(vertexShader->shader_data);
    free(vertexShader);

    free(pixelShader->shader_data);
    free(pixelShader);

    free(pixelShader->shader_data);
    free(pixelShader);

    free(attributes);
    free(fetchShader->shader_program);
    free(fetchShader);

    free(cubePosVtx);
    free(cubeColorBuf);

    //!*******************************************************************
    //!                         return to HBL                            *
    //!*******************************************************************
    memoryRelease();

    return EXIT_SUCCESS;
}

